-- |
-- Module      :  Composition.Sound.Keyboard
-- Copyright   :  (c) OleksandrZhabenko 2020-2021
-- License     :  MIT
-- Stability   :  Experimental
-- Maintainer  :  olexandr543@yahoo.com
--
-- Helps to create experimental music from a file (or its part) and a Ukrainian text. 
-- It can also generate a timbre for the notes. Uses SoX inside.

{-# OPTIONS_GHC -threaded #-}

module Composition.Sound.Keyboard (
  -- * Working with input and files
  qwerty2dvorak
  , dvorak2qwerty
  , input2BL
  , input2BLN
  , input2BLMN
  , readFile2BL
  , readFile2BLN
  , readFile2BLMN
  , readFile2BLGen  
  , readFile2BLGenN
  , readFile2BLGenMN
  -- * Conversions
  , readFileDoubles
  , readFileDoublesN
  , readFileDoublesMN
  , readFileDoublesGen
  , readFileDoublesGenN
  , readFileDoublesGenMN
  , takeDoubles
  , hashStr2
  , convH
) where

import CaseBi.Arr (getBFstL',getBFstLSorted')
import Data.Char (isAsciiLower)
import GHC.Arr
import Data.Foldable.Ix
import GHC.Int (Int64)

-- | Converts a lazy 'String' into a list of 'Int' using 'hashStr2'. 
takeDoubles :: String -> [Int]
takeDoubles xs = zipWith hashStr2 xs $ tail xs

-- | Converts a string of lowercase ASCII letters being typed on the QWERTY keyboard layout into corresponding Dvorak keyboard layout.
qwerty2dvorak :: String -> String
qwerty2dvorak = map (getBFstL' ' ' (zip ("/;<>" ++ ['a'..'z']) "wvszaxje.uidchtnmbrl'poygk,qf;"))

-- | Vice versa to 'qwerty2dvorak'.
dvorak2qwerty :: String -> String
dvorak2qwerty = map (getBFstL' ' ' (zip ("',.;" ++ ['a'..'z']) "qwezanihdyujgcvpmlsrxo;kf.,bt/"))

-- | Hashes two lower case ascii characters. Is used for controlling frequencies and operators.
hashStr2 :: Char -> Char -> Int
hashStr2 x y = getBFstLSorted' 57 (zip ['a'..'z'] . map (\t -> getBFstLSorted' (26 * t + 18) (zip ['a'..'z'] [(26 * t + 4)..]) y) . 
  concat $ [[0],[6..8],[1],[9..11],[4],[12..16],[2],[17..21],[3],[22..24],[5,25]]) x -- 679 is the greatest value ~ \"zz\"; there are 572 effectful val.

-- | Get contents into lazy 'String' with filtering of all characters that are not a lower case ascii letters.
input2BL :: IO (String)
input2BL = fmap (convH [] (filter isAsciiLower)) getContents

-- | Like 'input2BL', but takes only first @n@ symbols specified with the first 'Int' argument.
input2BLN :: Int -> IO (String)
input2BLN n = fmap (convH [] (take n . filter isAsciiLower)) getContents

-- | Like 'input2BL', but takes only first @n@ symbols specified with the second 'Int' argument dropping before this the first @m@ symbols specified 
-- with the first 'Int' argument.
input2BLMN :: Int -> Int -> IO (String)
input2BLMN m n = fmap (convH [] (take n . drop m . filter isAsciiLower)) getContents

-- | Reads a given file into a lazy 'String' with filtering of all characters that are not a lower case ascii letters. It has additional 
-- first command line argument to control the way of treating letters: as being typed (entered) properly (null 'String'), or needed to be converted 
-- from qwerty to dvorak layout (\"q\" 'String'), or vice versa (otherwise).
readFile2BLGen :: String -> FilePath -> IO (String)
readFile2BLGen ys = fmap (convH ys (filter isAsciiLower)) . readFile

-- | Like 'readFile2BLGen', but reads only first @n@ symbols specified with the first 'Int' argument.
readFile2BLGenN :: Int -> String -> FilePath -> IO (String)
readFile2BLGenN n ys = fmap (convH ys (take n . filter isAsciiLower)) . readFile

-- | Like 'readFile2BLGen', but reads only first @n@ symbols specified with the second 'Int' argument dropping before this the first @m@ symbols specified 
-- with the first 'Int' argument.
readFile2BLGenMN :: Int -> Int -> String -> FilePath -> IO (String)
readFile2BLGenMN m n ys = fmap (convH ys (take n . drop m . filter isAsciiLower)) . readFile

-- | Auxiliary function to define how is a 'String' treated, see 'readFile2BLGen'.
convH :: String -> (String -> String) -> (String -> String)
convH ys f 
 | null ys = f
 | ys == "q" = qwerty2dvorak . f
 | otherwise = dvorak2qwerty . f
  
-- | Usual way the function 'readFile2BLGen' is used. The text in a file being read is treated as a properly typed (entered) one. So there is no 
-- keyboard layout conversion at all.
readFile2BL :: FilePath -> IO (String)
readFile2BL = readFile2BLGen []
{-# INLINE readFile2BL #-}

-- | Like 'readFile2BL', but reads only first @n@ symbols specified with the first 'Int' argument.
readFile2BLN :: Int -> FilePath -> IO (String)
readFile2BLN n = readFile2BLGenN n []
{-# INLINE readFile2BLN #-}

-- | Like 'readFile2BL', but reads only first @n@ symbols specified with the second 'Int' argument dropping before this the first @m@ symbols specified 
-- with the first 'Int' argument.
readFile2BLMN :: Int -> Int -> FilePath -> IO (String)
readFile2BLMN m n = readFile2BLGenMN m n []
{-# INLINE readFile2BLMN #-}

-- | After reading a file into a filtered lazy 'String' (see, 'readFile2BLGen') converts the resulting 'String' into a list
-- of 'Int'. The arguments have the same meaning as for 'readFile2BLGen'.
readFileDoublesGen :: String -> FilePath -> IO [Int]
readFileDoublesGen ys = fmap (takeDoubles . convH ys (filter isAsciiLower)) . readFile

-- | Like 'readFileDoublesGen', but returns only first @n@ elements of the list specified with the first 'Int' argument.
readFileDoublesGenN :: Int -> String -> FilePath -> IO [Int]
readFileDoublesGenN n ys = fmap (take (fromIntegral n) . takeDoubles . convH ys (filter isAsciiLower)) . readFile

-- | Like 'readFileDoublesGen', but returns only first @n@ symbols specified with the second 'Int' argument dropping before this the first @m@ symbols specified 
-- with the first 'Int' argument.
readFileDoublesGenMN :: Int -> Int -> String -> FilePath -> IO [Int]
readFileDoublesGenMN m n ys = fmap (s2L (fromIntegral m) (fromIntegral n) . takeDoubles . convH ys (filter isAsciiLower)) . readFile

-- | Usual way the function 'readFileDoublesGen' is used. The text in a file being read is treated as a properly typed (entered) one. So there is no 
-- keyboard layout conversion at all.
readFileDoubles :: FilePath -> IO [Int]
readFileDoubles = readFileDoublesGen []
{-# INLINE readFileDoubles #-}

-- | Like 'readFileDoubles', but returns only first @n@ elements of the list specified with the first 'Int' argument.
readFileDoublesN :: Int -> FilePath -> IO [Int]
readFileDoublesN n = readFileDoublesGenN n []
{-# INLINE readFileDoublesN #-}

-- | Like 'readFileDoubles', but returns only first @n@ elements of the list specified with the second 'Int' argument 
-- dropping before this the first @m@ elements specified with the first 'Int' argument.
readFileDoublesMN :: Int -> Int -> FilePath -> IO [Int]
readFileDoublesMN m n = readFileDoublesGenMN m n []
{-# INLINE readFileDoublesMN #-}
